// MAIN program
// three passes:
// analyse
// transform_1
// transform_2

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" {
   void *malloc(...);
   int atoi(...);
   void free(...);
   void exit(...);
}
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "compilerApi.h"
#include "strategy.h"
#include "anls_att.h"

#include "indep.h"

extern void specification_analyse(SgStatement *s);

/*
extern void analyse_processor(SgStatement *);
extern void analyse_template(SgStatement *);
extern void analyse_dynamic(SgStatement *);
*/
extern void add_processor_declaration(SgStatement *);

extern void analyse_alignment(SgStatement *);
extern void analyse_realignment(SgStatement *);
extern void show_alignment(SgStatement *);
extern void add_template_declaration(SgStatement *);

extern void analyse_distribute(SgStatement *);
extern void analyse_redistribute(SgStatement *);
extern void redistribute_array(SgStatement *);
extern void show_distribute(SgStatement *);

/*
extern void analyse_inherit(SgStatement *);
extern void analyse_outofcore(SgStatement *);
extern void analyse_memory(SgStatement *);
extern void analyse_incoresize(SgStatement *);
extern void analyse_major(SgStatement *);
extern void analyse_shadow(SgStatement *);
*/
extern void analyse_common(SgStatement *);

extern void analyse_filemap(SgStatement *);
extern void trans_filemap(SgStatement *);

extern void trans_io_processor_stmt(SgStatement *);

extern void caculate_template_local_size(SgStatement *);

extern void add_align_array(SgStatement *);
extern void add_dummy_array(SgStatement *);
extern void realign_array(SgStatement *);

extern void analyse_forall_header(SgStatement *);
extern void get_forall_reference(SgStatement *);
extern void index_forall_reference(int, SgSymbol *, SgStatement *);
extern int detect_irregular_lhs(SgStatement *);
extern void detect_forall_reference(SgStatement *);
extern void show_forall_reference(SgStatement *);
extern void analyse_forall_template(SgStatement *);
extern void show_forall_template(SgStatement *);

extern void forall_do(SgStatement *);
extern int  scan_indo(SgStatement *);
extern void analyse_indo(SgStatement *);
extern void trans_indo(SgStatement *);
extern void trans_indo_call(SgStatement *);

extern void analyse_assign(SgStatement *);
extern void analyse_datatransfer(SgStatement *);
extern void analyse_proc_stat(SgStatement*);

extern void trans_stop(SgStatement *);
extern void trans_return(SgStatement *);
extern void analyse_if_stmt(SgStatement *);
extern void trans_if_stmt(SgStatement *);

extern void analyse_select_stmt(SgStatement *);

extern void analyse_on_stmt(SgStatement *);
extern void trans_on_stmt(SgStatement *);

extern void analyse_resident_stmt(SgStatement *);
extern void trans_resident_stmt(SgStatement *);

extern void analyse_dirend_stmt(SgStatement *s);
extern void trans_dirend_stmt(SgStatement *);

extern void analyse_task_stmt(SgStatement *s);
//------------------;

extern void define_return_label(SgStatement *);
extern void define_stop_label(SgStatement *);
extern void add_definition(SgStatement *);
extern void add_new_dec(SgStatement *);

extern void decide_global_shadow(SgStatement *);
extern void delete_align_array(SgStatement *);
extern void delete_allocatable_array(SgStatement *);
extern void delete_dummy_array(SgStatement *);

extern void trans_forall(SgStatement *);
extern int regularize_lhs_array(SgStatement *);

extern void trans_datacontrol(SgStatement *);
extern void trans_datatransfer(SgStatement *);

extern void trans_assign(SgStatement *);
extern void trans_pointer_assign(SgStatement *);

extern void trans_procstmt(SgStatement *);
extern void change_prochedr(SgStatement *);
extern void change_indepdo_prochedr(SgStatement *);

extern void process_label(SgStatement *s);

    //added by slx
extern void analyse_allocatable(SgStatement *);
extern void analyse_allocate(SgStatement * s);
extern void trans_allocate(SgStatement *);
extern void trans_deallocate(SgStatement *);
extern void transform_3(SgStatement *);

extern void extract_contains(SgStatement *);

extern void func_name_init(); 			// wjp

extern void indep_call_trans(SgStatement * s,int & proc_counter);

char * mixstring="pcrc";

//int stack_size = MAX_STACK_SIZE;
int stack_size = MIN_STACK_SIZE/4;
int heap_size = MAX_HEAP_SIZE;
int IN_CORE_SIZE = 250000;
int RECORD_SIZE = IN_CORE_SIZE;

SgFile * global_file;

//flags like the one in unparse;

Flag in_main_program = off;
Flag in_procedure = off;
Flag in_interface = off;
Flag in_module = off;
Flag in_independent = off;

int have_main_program = 0;

Flag profile_flag = off;
Flag clock_flag = off;

int optimize = DISABLE_OPTIMIZE;
int prefetchFlag = 0;

/* bit value meaning
0. INDEX_OPTIMIZE
1. LOCAL2GLOBAL
*/
int opt_strategy=0;

int debug_trans0 = 0;
int silentFlag = 0;

  // flag to pass to runtime library to enable debug mode
int runtimeDebug = 0;

int storageMode = COLUMN_MAJOR;
int ioMode = IO_FORTRAN;

int element_intrinsic_call = STRATEGY_ELEMENT_INTRINSIC_CALL;

int forall_counter = 0;
// record the level of forall .
char forall_level[2] ;
// record the level of expression .
char exp_level[1024] ;

//Communication {NONE, SHIFT, REMAP, BROADCAST_SCALAR, BROADCAST_DIM, IRREG, UNKNOW, BROADCAST};
int  comm_feq[15];

int do_level = 0;
int if_level = 0;

#define WAY 2

void puttab(int number)
{
  for (int i=0; i<number; i++)
    printf("  ");
}

void tree_exp(SgExpression * expr, int level)
{
  static int tabnumber=0;

  level--;
  tabnumber++;
  if(level!=0) {
    puttab(tabnumber);
    if (expr!=NULL)
      printf("id: %3d, variant: %3d\n", expr->id(), expr->variant());
    else {
      printf("nil nil nil nil nil \n");
      tabnumber--;
      return;
    }

    if (expr->operand(1)!=NULL) {
      tree_exp(expr->operand(1), level);
    }
    else {
      tabnumber++;
      puttab(tabnumber);
      printf("nil nil nil nil nil \n");
      tabnumber--;
    };

    if (expr->operand(2)!=NULL)
      tree_exp(expr->operand(2), level);
    else {
      tabnumber++;
      puttab(tabnumber);
      printf("nil nil nil nil nil \n");
      tabnumber--;
    };
  }
  tabnumber--;
}

SgExpression * find_exp(int id)
{
  SgExpression * temp = global_file->firstExpression();
  printf("Finding exp %3d ...\n", id);
  for (; temp; temp=temp->nextInExprTable()) {
    if(temp->id()==id) 
      break;
   }
 return temp;
}

void init_scope()
{
	forall_level[0]='\0';
	forall_level[1]='\0';
	strcpy(exp_level,"1");
/*
	exp_level[0]='a';
	exp_level[1]='\0';
*/
}

//------------------------;


int getreflevel(SgSymbol * index, SgExpression * expr)
{
  if (expr) {
    switch(expr->variant()) {

    case INT_VAL:
      return(0);
      break;

    case VAR_REF:
      if ((index)==(expr->symbol()))
	return (1);
      else
	return (0);
      break;
//add by huang
    case DDOT:
        return 0;
        break;
//add finished
    case MINUS_OP:
    case ADD_OP:    
      {
	int left, right;
	left=getreflevel(index, expr->operand(1));
	right=getreflevel(index, expr->operand(2));
	
	return((left>right) ? left:right);
	break;
      };

    case MULT_OP:      
      {
	int left, right;
	left=getreflevel(index, expr->operand(1));
	right=getreflevel(index, expr->operand(2));
	
	return((left+right));
	break;
      };

                // zheng.gb
//    case ARRAY_REF:
//	if (expr->IsSymbolInExpression(*index)) return 100;
//        else return 0;
//        break;

    case ARRAY_REF:
    case  FUNC_CALL:
    case  INTRIN_FUNC_CALL:
//      printf("getreflevel:function call.\n");
	// if this expression is index-invariant , and scalar type , see it as an invariant .
	if (!expr->IsSymbolInExpression(*index) && !isSgArrayType(expr->type())) return 0;
        else return -100; //100
        break;

    default:
      printf("getreflevel::Expression not touched: variant:%d id:%d\n", expr->variant(),expr->id());
      return (-100); //100
      break;
    };
  };

  return (-1);
}


// now we only have a simple to deal with s*index+o, and s!=0;
// so, i can only be 1 or 0;

SgExpression * getlinerexp(int i, SgSymbol * symb, SgExpression * expr)
{
  switch(expr->variant()) {
  case VAR_REF: {
      if (i==0) {
	return NULL;
      };
      if (i==1) {
	SgValueExp * stride = new SgValueExp(1);
	return stride;
      };
    break;
  };
   
  case MULT_OP: {
    if (i==0) {
      return NULL;
    };
    if (i==1) {
      int reflevel=getreflevel(symb, expr->operand(1));
      if (reflevel==1) 
	return expr->operand(2);
      else 
	return expr->operand(1);
    };
    break;
  }

  case ADD_OP: {
    if (i==0) {
      int reflevel=getreflevel(symb, expr->operand(1));
      if (reflevel==1) 
	return expr->operand(2);
      else 
	return expr->operand(1);
    };
    if (i==1) {
      int reflevel=getreflevel(symb, expr->operand(1));
      if (reflevel==1) 
	return (getlinerexp(1, symb, expr->operand(1)));
      else 
	return (getlinerexp(1, symb, expr->operand(2)));
    };
    break;
  }

  case  MINUS_OP: {
    if (i==0) {
      int reflevel=getreflevel(symb, expr->operand(1));
      if (reflevel==1) {
	SgExpression * minus = new SgExpression(MINUS_OP, NULL, NULL, NULL);
	minus->setLhs(expr->operand(2));
	return (minus);
      }
      else 
	return expr->operand(1);
    };
    if (i==1) {
      int reflevel=getreflevel(symb, expr->operand(1));
      if (reflevel==1) 
	return (getlinerexp(1, symb, expr->operand(1)));
      else {
	SgExpression * minus = new SgExpression(MINUS_OP, NULL, NULL, NULL);
	minus->setLhs(getlinerexp(1, symb, expr->operand(2)));
	return (minus);
      }
    };
    break;
  }
                // zheng.gb
  case  FUNC_CALL:
  case  INTRIN_FUNC_CALL:
        printf("getlinerexp:function call\n");
    if (i==0) {
      return NULL;
    };
        return 0;
  default:
    printf("Expression not touched!\n");
    break;
  };

  return NULL;
}

//------------------------;

SgExpression * forall_index(SgStatement * s)
{
  SgExpression * temp = s->expr(0);

  while (temp->variant()!=INDEX_OP)
    temp=temp->operand(1);
  return (temp->operand(1));
}

void show_template_from_symbol(SgSymbol * symb)
{
    int i=0;

    for (symb; symb; symb=symb->next()) {
      if(symb->numberOfAttributes(Temp_Def)) {
	((Template_def *)(symb->getAttribute(0, Temp_Def)->getAttributeData()))->show();
      };
    };
}


//This functionis not well implemented, we should use more sophiscated SgVarDec...;

void add_var_dec(SgStatement * s, SgType * type, SgSymbol * name)
{

  SgExpression * expr = new SgVarRefExp(name);
  
  SgExpression * exprtype = new SgExpression(TYPE_OP, NULL, NULL, NULL, type);
  
  SgStatement * declstatement =new SgStatement(VAR_DECL, NULL, NULL, expr, exprtype, NULL);

  s->childList1(0)->insertStmtBefore(*declstatement, *s);

  delete expr;
  delete exprtype;
  delete declstatement;
}

//--------------------------------;

char * int_typename="integer";
char * byte_typename="byte";
char * word_typename="word";
char * float_typename="real";
char * complex_typename="complex";
char * double_typename="double";
char * logical_typename="logical";
char * character_typename="character";

char * nameOfType(SgType * stype)
{    
  char * tname = NULL;

  if (stype->hasBaseType())
    stype=stype->baseType();
  ASSERT(stype);
  if (isSgKindType(stype))
    stype=stype->baseType();

  switch (stype->variant()) {
  case T_BOOL:
    tname=logical_typename;
    break;

  case T_CHAR:
  case T_STRING:
    tname=character_typename;
    break;

  case T_INT:
    tname=int_typename;
    break;
  case T_BYTE:
    tname=byte_typename;
    break;
  case T_WORD:
    tname=word_typename;
    break;
  case T_FLOAT:
    tname=float_typename;
    break;
  case T_DOUBLE:
    tname=double_typename;
    break;
  case T_COMPLEX:
    tname=complex_typename;
    break;
  case T_DCOMPLEX:
    
    break;
  case T_KIND:
  case T_ARRAY:
    ASSERT(0);
    break;
  default:
    // do nothing;
    break;
  };

  return tname;
}
//--------------------------------;

#ifdef MEM_ALLOC
char * int_allocname="integer_stack_alloc";
char * byte_allocname="byte_stack_alloc";
char * word_allocname="word_stack_alloc";
char * float_allocname="real_stack_alloc";
char * complex_allocname="complex_stack_alloc";
char * double_allocname="double_stack_alloc";
char * logical_allocname="logical_stack_alloc";
char * character_allocname="character_stack_alloc";
#else
char * int_allocname="integer_alloc";
char * byte_allocname="byte_alloc";
char * float_allocname="real_alloc";
char * complex_allocname="complex_alloc";
char * double_allocname="double_alloc";
char * logical_allocname="logical_alloc";
#endif

char * allocNameOfType(SgType * stype)
{    
  char * tname = NULL;

  if (stype->hasBaseType())
    stype=stype->baseType();
  if (isSgKindType(stype))
    stype=stype->baseType();

  switch (stype->variant()) {
  case T_BOOL:
    tname=logical_allocname;
    break;

  case T_CHAR:
  case T_STRING:
    tname=character_allocname;
    break;

  case T_INT:
    tname=int_allocname;
    break;
  case T_BYTE:
    tname=byte_allocname;
    break;
  case T_WORD:
    tname=word_allocname;
    break;
  case T_FLOAT:
    tname=float_allocname;
    break;
  case T_DOUBLE:
    tname=double_allocname;
    break;
  case T_COMPLEX:
    tname=complex_allocname;
    break;
  case T_DCOMPLEX:
    
    break;
  case T_ARRAY:
  case T_KIND:
    ASSERT(0);
    break;
  default:
    // do nothing;
    break;
  };

  return tname;
}

//--------------------------------;

#ifdef MEM_ALLOC
char * int_allocheapname="integer_heap_alloc";
char * byte_allocheapname="byte_heap_alloc";
char * word_allocheapname="word_heap_alloc";
char * float_allocheapname="real_heap_alloc";
char * complex_allocheapname="complex_heap_alloc";
char * double_allocheapname="double_heap_alloc";
char * logical_allocheapname="logical_heap_alloc";
char * character_allocheapname="byte_heap_alloc";
#else
char * int_allocheapname="integer_alloc";
char * byte_allocheapname="byte_alloc";
char * float_allocheapname="real_alloc";
char * complex_allocheapname="complex_alloc";
char * double_allocheapname="double_alloc";
char * logical_allocheapname="logical_alloc";
#endif

char * allocHeapNameOfType(SgType * stype)
{    
  char * tname = NULL;

  if (stype->hasBaseType())
    stype=stype->baseType();
  if (isSgKindType(stype))
    stype=stype->baseType();

  switch (stype->variant()) {
  case T_BOOL:
    tname=logical_allocheapname;
    break;

  case T_CHAR:
  case T_STRING:
    break;

  case T_INT:
    tname=int_allocheapname;
    break;
  case T_BYTE:
    tname=byte_allocheapname;
    break;
  case T_WORD:
    tname=word_allocheapname;
    break;
  case T_FLOAT:
    tname=float_allocheapname;
    break;
  case T_DOUBLE:
    tname=double_allocheapname;
    break;
  case T_COMPLEX:
    tname=complex_allocheapname;
    break;
  case T_DCOMPLEX:
    
    break;
  case T_ARRAY:
  case T_KIND:
    ASSERT(0);  
    break;
  default:
    // do nothing;
    break;
  };

  return tname;
}
//--------------------------------;

#ifdef MEM_ALLOC
char * freeNameOfType(SgType * stype)
{    
  char * tname = "mem_free";
  return tname;
}
#else
char * int_freename="integer_free";
char * byte_freename="byte_free";
char * float_freename="real_free";
char * complex_freename="complex_free";
char * double_freename="double_free";
char * logical_freename="logical_free";
char * char_freename="character_free";

char * freeNameOfType(SgType * stype)
{    
  char * tname = NULL;

  if (stype->hasBaseType())
    stype=stype->baseType();

  switch (stype->variant()) {
  case T_BOOL:
    tname=logical_freename;
    break;

  case T_CHAR:
  case T_STRING:
    tname=character_free;
    break;

  case T_INT:
    tname=int_freename;
    break;
  case T_BYTE:
    tname=byte_freename;
    break;
  case T_FLOAT:
    tname=float_freename;
    break;
  case T_DOUBLE:
    tname=double_freename;
    break;
  case T_COMPLEX:
    tname=complex_freename;
    break;
  case T_DCOMPLEX:
    
    break;
  case T_ARRAY:
    
    break;
  default:
    // do nothing;
    break;
  };

  return tname;
}
#endif

//--------------------------------;

char * int_typesizename="size_integer";
char * float_typesizename="size_real";
char * byte_typesizename="size_byte";
char * word_typesizename="size_word";
char * complex_typesizename="size_complex";
char * double_typesizename="size_double";
char * logical_typesizename="size_logical";
char * char_typesizename="size_character";

char * sizeNameOfType(SgType * stype)
{    
  char * tname = NULL;

  if (stype->hasBaseType())
    stype=stype->baseType();
  if (isSgKindType(stype))
    stype=stype->baseType();

  switch (stype->variant()) {
  case T_BOOL:
    tname=logical_typesizename;
    break;

  case T_CHAR:
  case T_STRING:
    tname=char_typesizename;
    break;

  case T_INT:
    tname=int_typesizename;
    break;
  case T_BYTE:
    tname=byte_typesizename;
    break;
  case T_WORD:
    tname=word_typesizename;
    break;
  case T_FLOAT:
    tname=float_typesizename;
    break;
  case T_DOUBLE:
    tname=double_typesizename;
    break;
  case T_COMPLEX:
    tname=complex_typesizename;
    break;
  case T_DCOMPLEX:
    
    break;
  case T_ARRAY:
  case T_KIND:
    ASSERT(0);  
    break;
  default:
    // do nothing;
    break;
  };

  return tname;
}
//--------------------------------;

#ifdef MEM_ALLOC
char * int_typestackname="int";
char * byte_typestackname="byte";
char * word_typestackname="word";
char * float_typestackname="real";
char * complex_typestackname="scpl";
char * double_typestackname="dbl";
char * logical_typestackname="log";
char * character_typestackname="character";
#else
char * int_typestackname="integer_stack";
char * byte_typestackname="byte_stack";
char * float_typestackname="real_stack";
char * complex_typestackname="complex_stack";
char * double_typestackname="double_stack";
char * logical_typestackname="logical_stack";
#endif

char * stackNameOfType(SgType * stype)
{    
  char * tname = NULL;

  if (stype->hasBaseType())
    stype=stype->baseType();
  if (isSgKindType(stype))
    stype=stype->baseType();

  switch (stype->variant()) {
  case T_BOOL:
    tname=logical_typestackname;
    break;

  case T_CHAR:
  case T_STRING:
    tname=character_typestackname;
    break;

  case T_INT:
    tname=int_typestackname;
    break;
  case T_BYTE:
    tname=byte_typestackname;
    break;
  case T_WORD:
    tname=word_typestackname;
    break;
  case T_FLOAT:
    tname=float_typestackname;
    break;
  case T_DOUBLE:
    tname=double_typestackname;
    break;
  case T_COMPLEX:
    tname=complex_typestackname;
    break;
  case T_DCOMPLEX:
    
    break;
  case T_ARRAY:
  case T_KIND:
    ASSERT(0);  
    break;
  default:
    // do nothing;
    break;
  };

  return tname;
}

//-----------------------------------------;

//add by huang
void analyse(SgStatement * s);

void change_header(SgStatement * s){
        SgExpression * doHeaderExp=s->expr(0);
	if (doHeaderExp)
	{
        SgExpression * lboundExp=doHeaderExp->lhs()->lhs();
        SgExpression * uboundExp=doHeaderExp->lhs()->rhs();
        SgExpression * stepExp=doHeaderExp->rhs();
        collect_array_ref(s,lboundExp,0);
        collect_array_ref(s,uboundExp,1);
        collect_array_ref(s,stepExp,1);
        analyse_array_ref_list(s,s);
	}
}
//add finished

/**@name Analyse
 *
 */
void analyse(SgStatement * s)
{
  switch(s->variant()) {

  case GLOBAL:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);
      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      break;
    };

  case PROG_HEDR:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i; 

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      // do nothing?;
      in_main_program=on;

      init_scope();

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      in_main_program=off;

      have_main_program = 1;

      break;

    };

  case PROC_HEDR:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

	Flag kept;
	if(in_interface==on) kept = in_procedure;
      in_procedure=on;
      init_scope();

      for (i=0; i<prochedr->numberOfParameters(); i++) {
	prochedr->parameter(i)->addAttribute(DUMP_Def, NULL, 0);
	prochedr->addAttribute(DUMP_Def, NULL, 0);
      }

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      in_procedure = off;
	if(in_interface==on) in_procedure=kept;

      break;
    };

  case FUNC_HEDR:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;
      init_scope();

      for (i=0; i<prochedr->numberOfParameters(); i++) {
	prochedr->parameter(i)->addAttribute(DUMP_Def, NULL, 0);
	prochedr->addAttribute(DUMP_Def, NULL, 0);
      }

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      in_procedure = off;

      break;
    };

  case INTERFACE_STMT:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_interface=on;

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      in_interface = off;

      break;
    };

  case MODULE_STMT:
    {
#ifndef USE_F90
printf("Warning: MODULE is not supported .\n");
#else
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i; 

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      // do nothing?;
      in_module=on;

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      in_module=off;
#endif
    break;
    }

  case PROCESSORS_STMT:
//    analyse_processor(s);
    break;

  case TEMPLATE_STMT:
//    analyse_template(s);
    break;

  case DYNAMIC_DECL:
//    analyse_dynamic(s);
    break;

  case ALIGN_DECL:
    analyse_alignment(s);
    break;
    
  case REALIGN_STMT:
    analyse_realignment(s);
    break;
    
  case DISTRIBUTE_DECL:
    analyse_distribute(s);
    break;

  case REDISTRIBUTE_STMT:
    analyse_redistribute(s);
    break;

  case INHERIT_DECL:
//    analyse_inherit(s);
    break;

  case OUTOFCORE_DECL:
//    analyse_outofcore(s);
    break;

  case MEMORY_DECL:
//    analyse_memory(s);
    break;

  case INCORESIZE_DECL:
//    analyse_incoresize(s);
    break;

  case MAJOR_DECL:
//    analyse_major(s);
    break;

  case FILEMAP_STMT:
    analyse_filemap(s);
    break;

  case IO_PROCESSOR_STMT:
    break;

  case SHADOW_STMT:
//    analyse_shadow(s);
    break;

  case ALLOCATABLE_STMT:
    break;

  case ALLOCATE_STMT:
        //slx
#ifndef USE_f90
    analyse_allocate(s);
#endif
    break;

  case DEALLOCATE_STMT:
    break;

  case VAR_DECL:
    // slx
    analyse_allocatable(s);
    break;

  case COMM_STAT:
    analyse_common(s);
    break;

  case EXTERN_STAT:
    // do nothing
    break;

  case CONTAINS_STMT:
    break;

  case INDEPENDENT_STMT:        //lan
    in_independent = on;  //or s->lexNext()->addAttribute(INDEP_Def...)?
    break;

  case FOR_NODE:
    {
      do_level++;

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      if(in_independent == on) {      //lan
// added by huang
printf("........Independent do statement analyse begin..........\n");
	   analyse_indepdo_header(s);
           get_indepdo_reference(s);
           detect_indepdo_reference(s);
           display_indepdo_reference_communication(s);
printf("........Independent do statement analyse end..........\n");
//

#if ( WAY == 1 )
	   //parse: to do or not to do
	   if(scan_indo(s)==1) {

                in_independent = off;

      		for (i=0; i<org_numberOfChildren; i++)
			if(children[i]->variant() == FORALL_STMT)
				forall_do(children[i]);


                s->addAttribute(INDEP_Def, NULL, 0);
                for (i=0; i<org_numberOfChildren; i++)
                        children[i]->addAttribute(INDEP_Def, NULL, 0);

                analyse_indo(s);
	   }
	   else {
                in_independent = off;
      		for (i=0; i<org_numberOfChildren; i++)
			analyse(children[i]);
	   }
#else
                in_independent = off;
#endif
      }
      else {
      		for (i=0; i<org_numberOfChildren; i++)
			analyse(children[i]);
      }

//      for (i=0; i<org_numberOfChildren; i++)
//	analyse(children[i]);
      
      delete [] children;
      
      do_level--;

      break;
    }

  case DO_WHILE_NODE:
    {
      do_level++;

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      do_level--;
     break;
    }

  case LOGIF_NODE:
  case LOGIF_STMT:
    {
      analyse_if_stmt(s);	// by zheng.gb

      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int org_numberOfChildren2=s->numberOfChildrenList2();

      SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	analyse(children1[i]);
      
      delete [] children1;

      for (i=0; i<org_numberOfChildren2; i++)
	children2[i]=s->childList2(i);

      for (i=0; i<org_numberOfChildren2; i++)
	analyse(children2[i]);
      
      delete [] children2;

      break;
    }

  case ELSEIF_NODE:
    {
      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int org_numberOfChildren2=s->numberOfChildrenList2();

      SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	analyse(children1[i]);
      
      delete [] children1;

      for (i=0; i<org_numberOfChildren2; i++)
	children2[i]=s->childList2(i);

      for (i=0; i<org_numberOfChildren2; i++)
	analyse(children2[i]);
      
      delete [] children2;

      break;
    }

  case SWITCH_NODE:
	analyse_select_stmt(s);
  case CASE_NODE:
    {
      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	analyse(children1[i]);
      
      delete [] children1;

      break;
    }

  case SIMPLE_ON_STMT:
      break;

  case BLOCK_ON_STMT: {

      analyse_on_stmt(s);

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
    }
    break;

  case SIMPLE_RESIDENT_STMT:
      break;

  case BLOCK_RESIDENT_STMT: {

      analyse_resident_stmt(s);

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
    }
    break;

  case TASK_STMT: {
      analyse_task_stmt(s);

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
    }
    break;

  case DIR_END:
    analyse_dirend_stmt(s);
    break;

  case ASSGOTO_NODE:
    // do nothing;
    break;

  case POINTER_ASSIGN_STAT:
    // do nothing;
    break;

  case ASSIGN_STAT:
    intrinsic2sub(s);
    if (s->variant() == ASSIGN_STAT ) {
      analyse_assign(s);
      break;
    }
    // else continue process 

  case PROC_STAT:
    //do nothing at presenttime. we will analyse INTERFACE later;
    if(in_procedure==on){//added by huang
	SgStatement *s1=s->controlParent();
        int variant=s1->variant();
        while(variant!=PROC_HEDR){
                s1=s1->controlParent();
                variant=s1->variant();
        }

	ASSERT(s1);

	SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s1;

	if(s1->symbol()->numberOfAttributes(CALLED_IN_INDEP)){
	SgCallStmt * call=(SgCallStmt *)s;
        SgFunctionCallExp *proc_call = (SgFunctionCallExp *)call->expr(0);
	SgSymbol *proc_name=proc_call->symbol();
	SgFunctionSymb *procname=(SgFunctionSymb *)proc_call->rhs()->symbol();
	int has_array_arg=0;
	for(int i=0;i<procname->numberOfParameters();i++){
		if(procname->parameter(i)->type()->variant()==T_ARRAY){
			has_array_arg=1;
			break;
		}
	}
	if(has_array_arg) proc_name->addAttribute(CALLED_IN_INDEP,NULL,0);
	}
	

    }
    analyse_proc_stat(s);
    break;

  case FORALL_STMT:
    {       
      analyse_forall_header(s);
      get_forall_reference(s);

      Forall_header_def * fhdef=(Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();
    
      for (int i=0; i<fhdef->getindexnumber(); i++) {
	SgSymbol * index=fhdef->getindexsymb(i);
	index_forall_reference(i, index, s->childList1(0));
      };

      // regular lhs array
      if (regularize_lhs_array(s)) { analyse(s); break; }

      // make the index seq in forall header same as lhs array ref
//      if (opt_strategy & OTHER_OPTIMIZE)
        // get LHS's storage mode
      int storageStratgy = decide_optimize_major(s);
      if (storageStratgy == COLUMN_MAJOR)
      {
           if (column_major(s)) { 
		analyse(s); break; 
	   }
      }
      else {
           if (row_major(s)) { 
		analyse(s); break; 
	   }
      }

      // detect reference's communication pattern
      detect_forall_reference(s);

      break;
    }

  case OPEN_STAT:
  case CLOSE_STAT:
// do nothing
    break;

  case READ_STAT:
  case PRINT_STAT:
  case WRITE_STAT:
    analyse_datatransfer(s);
    break;

  case CONTROL_END:
  case PARAM_DECL:
  case CONTINUE_NODE:
  case ENDDO_STMT:
  case  FORMAT_STAT:
   // statement need do nothing;
    break;

  case STOP_STAT: {
    SgStatement *scope = s->getScopeForDeclare();
    if (scope->variant() == PROG_HEDR)
    	scope->addAttribute(RETURN_LABEL, NULL, 0);
    break;
  }

  case RETURN_STAT: {
    SgStatement *scope = s->getScopeForDeclare();
    scope->addAttribute(RETURN_LABEL, NULL, 0);
    break;
   }

  case EXIT_STMT:
  case IMPL_DECL:
  case USE_STMT:
   // statement need do nothing;
   break;

  default:
    if (!silentFlag)
    printf("Not touched statement in analyse on line %d,variant:%d\n", s->lineNumber(),s->variant());
    break;
  };

}

void transform_1(SgStatement * s) 
{
   // keep label as continue statement.
  process_label(s);

  switch(s->variant()) {
  case GLOBAL:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];
      
      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);
      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);
      
      delete [] children;
      break;
    };

  case PROG_HEDR:
    {
      in_main_program = on;

if (!silentFlag)
printf("\n**** %s (main) === BEGIN Transformation ===\n",s->symbol()->identifier());

      // first remember the thing to do;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      init_scope();

      // need interface statement
      add_processor_declaration(s);

        // define exit label if needed.
      define_stop_label(s);

      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);

      add_template_declaration(s);

      caculate_template_local_size(s);

      decide_global_shadow(s);
      
      add_align_array(s);                  //   VAR_DECL of array

      //show_processor(s);
      //show_template(s);
      
      add_definition(s);
      
      delete [] children;
      
      in_main_program=off;

      break;
    }

  case PROC_HEDR:
    {
if (!silentFlag)
printf("\n**** %s (procedure) === BEGIN Transformation ===\n",s->symbol()->identifier());
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;
      init_scope();

      if (prochedr->name().numberOfAttributes(INDEP_Def) > 0) {
if (!silentFlag)
printf("IN INDEP_Def\n");
	trans_indo_call(s);
        add_dummy_array(s);
	break;

      }
     else if(s->symbol()->numberOfAttributes(CALLED_IN_INDEP)) {//added by huang 99.6.22
		printf("...Start to implement a serialized version of this subroutine ...\n");
		int proc_counter=0;
		SgStatement *s1=s->copyPtr();
		SgFunctionSymb *newproc=(SgFunctionSymb *)s1->getSymbolInCurrentScope(s1->symbol()->identifier(),"indepdo",NULL);
		newproc->setVariant(PROCEDURE_NAME);

		for(i=0;i<prochedr->numberOfParameters();i++){
			newproc->addParameter(i,*prochedr->parameter(i));
		}
		s1->replaceSymbBySymb(*s1->symbol(),*newproc);
		s->insertStmtBefore(*s1,*s->controlParent());
		indep_call_trans(s1,proc_counter);
      		change_indepdo_prochedr(s1);
      		add_dummy_array(s1);
		printf("...Done,The name of the serialized version is %s...\n",newproc->identifier());
		
	}

      add_processor_declaration(s);
     
        // define exit label if needed.
      define_return_label(s);

      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);

      add_template_declaration(s); 

      caculate_template_local_size(s);

      decide_global_shadow(s);	// zheng.gb

      change_prochedr(s);

/*
	add_definition < add_align_array:  add_definition will add push, pop ; while add_align_array may add dummy to header
*/
      add_definition(s);

      add_align_array(s);
      add_dummy_array(s);

//      add_definition(s);
      
//      change_prochedr(s);


      delete [] children;
      
      in_procedure = off;

      break;
    };

  case FUNC_HEDR:
    {
if (!silentFlag)
printf("\n**** %s (function) === Begin Transformation ===\n",s->symbol()->identifier());
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;
      init_scope();

      add_processor_declaration(s);

        // define exit label if needed.
      define_return_label(s);

      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);

      add_template_declaration(s); 
     
      caculate_template_local_size(s);

      decide_global_shadow(s);	// zheng.gb

      change_prochedr(s);

      add_definition(s);

      add_dummy_array(s);
      add_align_array(s);

//      add_definition(s);
      
//      change_prochedr(s);

      delete [] children;
      
      in_procedure = off;
 
      break;
    };

  case MODULE_STMT:
    {
      in_module = on;

if (!silentFlag)
printf("\n**** %s (module) === BEGIN Transformation ===\n",s->symbol()->identifier());

#ifndef USE_F90
printf("Warning: MODULE transformation ignored.\n");
    s->deleteStmt();
#else

      // first remember the thing to do;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      init_scope();

      // need interface statement
      add_processor_declaration(s);

      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);

      add_template_declaration(s);

      caculate_template_local_size(s);

      decide_global_shadow(s);
      
      add_align_array(s);                  //   VAR_DECL of array

      //show_processor(s);
      //show_template(s);
      
      add_definition(s);
      
      delete [] children;
      
#endif
      in_module=off;

      break;
    }

  case CONTAINS_STMT:
    break;

  case INTERFACE_STMT:
    s->deleteStmt();
    break;

  case PROCESSORS_STMT:
    //printf("- Processor Statement's id: % 4d variant:  % 4d\n", s->id(), s->variant());
    s->deleteStmt();
    break;

  case TEMPLATE_STMT:
    //printf("- Template Statement's  id: % 4d variant:  % 4d\n", s->id(), s->variant());
    s->deleteStmt();
    break;

  case DYNAMIC_DECL:
    s->deleteStmt();
    break;

  case ALIGN_DECL:
    //printf("- Align Statement's     id: % 4d variant:  % 4d\n", s->id(), s->variant());
    s->deleteStmt();
    break;

  case REALIGN_STMT:
    realign_array(s);
    break;

  case DISTRIBUTE_DECL:
    s->deleteStmt();
    break;

  case REDISTRIBUTE_STMT:
    redistribute_array(s);
    break;

  case INHERIT_DECL:
    s->deleteStmt();
    break;

  case OUTOFCORE_DECL:
  case INCORESIZE_DECL:
  case MEMORY_DECL:
  case MAJOR_DECL:
    s->deleteStmt();
    break;

  case FILEMAP_STMT:
    trans_filemap(s);
    break;

  case IO_PROCESSOR_STMT:
    trans_io_processor_stmt(s);
    break;

  case SHADOW_STMT:
    s->deleteStmt();
    break;

  case VAR_DECL:
    delete_align_array(s); 
    delete_allocatable_array(s);

    if (in_procedure==on)
      delete_dummy_array(s);

    break;

  case INDEPENDENT_STMT:
    s->deleteStmt();
    break;

  case FOR_NODE:
    {
      do_level++;

//add by huang
      change_header(s);
//add finished!

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

        //add the indo transformation if there is INDEP_Def
#if (WAY == 1)
// done by LAN, which is currently not used
      if(s->numberOfAttributes(INDEP_Def)) {
          trans_indo(s);
// huang
          if(s->numberOfAttributes(INDEPDO_Def)>0){
                delete_indep_stmt(s);
                dispose_indepdo_oc_reference(s);
          }
      }
#else
// done by HUANG
      if(s->numberOfAttributes(INDEPDO_Def)>0){
printf("........Independent do statement transformation begin..........\n");
                delete_indep_stmt(s);
                if(s->numberOfAttributes(Indepdo_Lref)>0){
                        change_indepdo_body(s);
                }
printf("........Independent do statement transformation end..........\n");

      }
#endif
      else {
          for (i=0; i<org_numberOfChildren; i++)
	    transform_1(children[i]);
      }
      
      delete [] children;
      
      do_level--;

      break;
    }

  case DO_WHILE_NODE:
    {
      do_level++;

      SgExpression * doHeaderExp=s->expr(0);
      analyse_expression(s, doHeaderExp);

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
        transform_1(children[i]);
      
      delete [] children;
      
      do_level--;

      break;
    }

  case LOGIF_NODE:
  case ELSEIF_NODE:
  case LOGIF_STMT:
    {
      if_level ++;

      trans_if_stmt(s);		// zheng.gb

      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int org_numberOfChildren2=s->numberOfChildrenList2();

      SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	transform_1(children1[i]);
      
      delete [] children1;

      for (i=0; i<org_numberOfChildren2; i++)
	children2[i]=s->childList2(i);

      for (i=0; i<org_numberOfChildren2; i++)
	transform_1(children2[i]);
      
      delete [] children2;

	// LOGIF_STMT => LOGIF_NODE
      if (s->variant() == LOGIF_STMT)
//	  (s->numberOfChildrenList1() > 2 || s->childList1(0)->numberOfChildrenList1() >1 ) )
      {
	s->setVariant(LOGIF_NODE);
      }

      if_level --;

      break;
    }

  case SWITCH_NODE: {
      if_level ++;

      trans_if_stmt(s);
    }
  case CASE_NODE:
    {
      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int org_numberOfChildren2=s->numberOfChildrenList2();

      SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	transform_1(children1[i]);
      
      delete [] children1;

      for (i=0; i<org_numberOfChildren2; i++)
	children2[i]=s->childList2(i);

      for (i=0; i<org_numberOfChildren2; i++)
	transform_1(children2[i]);
      
      delete [] children2;

      if (s->variant() == SWITCH_NODE )if_level --;

      break;
    }

  case SIMPLE_ON_STMT:
      break;

  case BLOCK_ON_STMT: {

      trans_on_stmt(s);

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);
      
      delete [] children;
    }
    break;

  case SIMPLE_RESIDENT_STMT:
      break;

  case BLOCK_RESIDENT_STMT: {

      trans_resident_stmt(s);

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);
      
      delete [] children;
    }
    break;

  case TASK_STMT: {

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);
      
      delete [] children;
    }
    break;

  case ASSGOTO_NODE:
    // do nothing;
    break;

  case STOP_STAT:
    trans_stop(s);
    break;

  case PROC_STAT:
    trans_procstmt(s);
    break;

  case ASSIGN_STAT:
    trans_assign(s);
    break;

  case POINTER_ASSIGN_STAT:
    trans_pointer_assign(s);
    break;

  case FORALL_STMT:
    //show_forall_reference(s);
    trans_forall(s);
    s->deleteStmt();
    break;

  case OPEN_STAT:
  case CLOSE_STAT:
    trans_datacontrol(s);
    s->deleteStmt();
    break;

  case READ_STAT:
  case PRINT_STAT:
  case WRITE_STAT:
    trans_datatransfer(s);
    s->deleteStmt();
    break;

  case DIR_END:
    trans_dirend_stmt(s);
    break;

  case CONTROL_END:
  case CONTINUE_NODE:
  case PARAM_DECL:
  case ENDDO_STMT:
  case  FORMAT_STAT:
    // statement need do nothing;
    break;

  case RETURN_STAT:
    trans_return(s);
   break;

  case INTRIN_STAT:
    s->deleteStmt();
   break;

  case EXIT_STMT:
  case IMPL_DECL:
  case EXTERN_STAT:
  case COMM_STAT:
   // statement need do nothing;
   break;

  case ALLOCATABLE_STMT:
    s->deleteStmt();
    break;

  case ALLOCATE_STMT:
        //slx
#ifndef USE_f90
    trans_allocate(s);
#endif
    break;

  case DEALLOCATE_STMT:
        //slx
#ifndef USE_f90
    trans_deallocate(s);
#endif
    break;

  default:
    if (!silentFlag)
    printf("In transform_1, unexpected statement on line %d, id: %d, variant: %d\n", s->lineNumber(), s->id(), s->variant());
    break;
  };
}

// add_new_dec in contain list
void deal_contains(SgStatement * s)
{
      int no;
      SgStatement *contains = get_contains(s, no);
      if (contains) {
	for (int i=no+1; i<s->numberOfChildrenList1()-1 ; i++) 
	{
	  SgStatement *stmt = s->childList1(i);
	  if (stmt->variant() == PROC_HEDR || stmt->variant() == FUNC_HEDR)
	      add_new_dec(stmt);
	}
#ifndef USE_F90
        extract_contains(s);
#endif
      }
}

void transform_2(SgStatement * s) 
{
  switch(s->variant()) {
  case GLOBAL:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);
      for (i=0; i<org_numberOfChildren; i++)
	transform_2(children[i]);
      
      delete [] children;

      break;
    };

  case PROG_HEDR:
    {
      in_main_program = on;
      add_new_dec(s);
      in_main_program = off;

      deal_contains(s);
      break;
    };

  case PROC_HEDR:
    {
	if(s->symbol()->numberOfAttributes(CALLED_IN_INDEP)) {//added by huang 99.6.22
//		indep_call_trans(s);
      in_procedure = on;
      add_new_dec(s);      
      in_procedure = off;
      deal_contains(s);
		break;
	}
      in_procedure = on;
      add_new_dec(s);      
      in_procedure = off;

      deal_contains(s);
      break;
    };

  case FUNC_HEDR:
    {
      in_procedure = on;
      add_new_dec(s);      
      in_procedure = off;

      deal_contains(s);
      break;
    };

  case MODULE_STMT:
    {
#ifdef USE_F90
      in_module = on;
      add_new_dec(s);      
      in_module = off;
#endif

      deal_contains(s);
      break;
    }

  default:
    if (!silentFlag)
    printf("Not touched statement in transform_2 on line %d variant:%d.\n", s->lineNumber(),s->variant());
    break;
  };
}

void test_transform(SgStatement * s) 
{
  while (s!=NULL) {
    if (s->variant()==ASSIGN_STAT) {
	printf("The right is int %d, %d\n",s->expr(1)->isInteger(), s->expr(1)->valueInteger());
    }
    
    s=s->lexNext();
  }
  
  s=global_file->firstStatement();

  SgLabel * label=global_file->firstLabel();

  while (label!=NULL) {
    printf("the lable id is: %d, %d\n", label->id(), label->labelValue());
    label=label->nextInScope(s);
  }
}

char *newOutputFileName(char *inputFileName)
{	
  char	*out;
  int	nameLength = strlen(inputFileName);
  
  out=(char *)malloc((unsigned)(nameLength+5));

  while (inputFileName[nameLength-1]!='.') {
    nameLength --;
  }

  for ( int i = 0; i<nameLength; i++) {
    out[i]=inputFileName[i];
  }

  out[nameLength-1]='\0';

  out=strcat(out,"_0.dep");

  return(out);
}

void display_comm_feq()
{
//Communication {NONE, SHIFT, REMAP, BROADCAST_SCALAR, BROADCAST_DIM, IRREG, UNKNOW, BROADCAST};
    printf("===============================\n");
    printf("The summary of communication:\n");
    for (int i=0;i<15;i++)
    {
	if (comm_feq[i] == 0) continue;
	switch (i) {
	case 0: printf("NONE:"); break;
	case 1: printf("SHIFT:"); break;
	case 2: printf("REMAP:"); break;
	case 3: printf("BROADCAST_SCALAR:"); break;
	case 4: printf("BROADCAST_DIM:"); break;
	case 5: printf("IRREG:"); break;
	case 6: printf("UNKNOW:"); break;
	case 7: printf("BROADCAST:"); break;
	}
	printf("  %d\n",comm_feq[i]);
    }
}

void init_compiler()
{
	func_name_init(); 			// wjp

// set optimize bit value as default;
	opt_strategy |= INDEX_OPTIMIZE;
//	opt_strategy |= LOCAL2GLOBAL;
	opt_strategy |= RUNTIMECOMM;

	if (debug_trans0) for (int i =0; i<15 ; i++) comm_feq[i]=0;
}

void analyse_opt(char *arg,int &i)
{
	if (arg[i]=='\0') {
		opt_strategy |= INDEX_OPTIMIZE;
		opt_strategy |= LOCAL2GLOBAL;
		opt_strategy |= RUNTIMECOMM;
		opt_strategy |= OTHER_OPTIMIZE;
	}
	for (; arg[i]!='\0' ; i++)
	{
		switch (arg[i]) {
		case '1':  opt_strategy ^= INDEX_OPTIMIZE; break;
		case '2':  opt_strategy ^= LOCAL2GLOBAL; break;
		case '3':  opt_strategy ^= RUNTIMECOMM; break;
		default:  break;
		}
	}
	i--;
}

static void usage()
{
        printf("Usage: trans0 [-h|-help] [-p] [-t] [-O|1|2|3] [-d] [-S in-core-size]\n");
        printf("        [-M memory] [-R record-size] [-D] HPF_file.dep \n");
	printf("options:\n");
        printf("       -h    : usage          \n");
        printf("       -p    : profile enabled . \n");
        printf("       -t    : pcrc_clock() is added for forall statement.\n");
        printf("       -O    : all optimization performed.\n");
        printf("       -O1   : optimize the local array subscripts calculation.\n");
        printf("       -O2   : optimize the local index to global index conversion..\n");
        printf("       -O3   : optimize the runtime communication detection.\n");
        printf("       -d    : debug mode enabled .\n");
        printf("       -S    : specify IN CORE ARRAY size .\n");
        printf("       -M    : specify dynamic stack size .\n");
        printf("       -R    : sepecify fortran file record size .\n");
        printf("       -D    : Enable runtime debug mode.\n");
}

int get_size(char *csize)
{
      	int size;
	int coef = 1;
	char last = csize[strlen(csize)-1];
	if (last == 'M' || last == 'm') 
	{
		coef = 1000000;
		csize[strlen(csize)-1] = '\0';
	}
	else if (last == 'K' || last == 'k') 
	{
		coef = 1000;
		csize[strlen(csize)-1] = '\0';
	}
	return atoi(csize) * coef;
}

void parse_string(char *s)
{
	if (strncmp(s,"major=",6) == 0) {
		switch (s[6]) {
		case 'c':
			storageMode = COLUMN_MAJOR; break;
		case 'r':
			storageMode = ROW_MAJOR; break;
		default:
        		printf("warning : Unknown option \"%s\" ignored.\n",s);
		}
	}
	else 
     	if (strncmp(s,"io=",3) == 0) {
		switch (s[3]) {
		case 'c':
			ioMode = IO_C; break;
		case 'f':
			ioMode = IO_FORTRAN; break;
		default:
        		printf("warning : Unknown option \"%s\" ignored.\n",s);
		}
	}
        else
     	if (strncmp(s,"prefetch",8) == 0) {
		prefetchFlag = 1;
	}
        else
     	if (strncmp(s,"help",4) == 0) {
		usage();
		exit(0);
	}
	else
       		printf("warning : Unknown option \"%s\" ignored.\n",s);
}

main(int argc, char * argv[]) 
{
	int i;

        init_compiler();

	argv++;
        argc--;
	if (argc == 0) { usage(); exit(1); }
        while ((argc > 0) && ((*argv)[0] == '-') )
        {
                i=1;
                while (argv[0][i] != '\0') {
                        switch (argv[0][i]) {
                        case 'p'  :
				profile_flag=on;
				break;
                        case 't'  :
				clock_flag=on;
				break;
                        case 'O'  :
				i++;
				analyse_opt(argv[0],i);
				break;
                        case 'd'  :
				debug_trans0=1;
				break;
                        case 'D'  :
				runtimeDebug=1;
				break;
                        case 'S'  :
				if (argv[0][i+1]!='\0') {
					IN_CORE_SIZE = get_size(argv[0]+i+1);
                                        i=strlen(argv[0]);
                                        continue;
                                }
                                else {
                                   if (argc == 1) {
                                        printf("warning : Invalid option \"-S\" ignored.\n");
                                   }
                                   else {
                                        argc--;
                                        argv++;
					IN_CORE_SIZE = get_size(argv[0]);
                                        i=strlen(argv[0])-1;   /* quit loop */
                                   }
                                }
                                break;
                        case 'M'  :
				if (argv[0][i+1]!='\0') {
					stack_size = -get_size(argv[0]+i+1);
                                        i=strlen(argv[0]);
                                        continue;
                                }
                                else {
                                   if (argc==1) {
                                        printf("warning : Invalid option \"-M\" ignored.\n");
                                   }
                                   else {
                                        argc--;
                                        argv++;
					stack_size = -get_size(argv[0]);
                                        i=strlen(argv[0])-1;   /* quit loop */
                                   }
                                }
                                break;
                        case 'R'  :
				if (argv[0][i+1]!='\0') {
					RECORD_SIZE = get_size(argv[0]+i+1);
                                        i=strlen(argv[0]);
                                        continue;
                                }
                                else {
                                   if (argc==1) {
                                        printf("warning : Invalid option \"-R\" ignored.\n");
                                   }
                                   else {
                                        argc--;
                                        argv++;
					RECORD_SIZE = get_size(argv[0]);
                                        i=strlen(argv[0])-1;   /* quit loop */
                                   }
                                }
                                break;
                        case 'z'  :
				silentFlag=1;
				break;
			case 'h'  :
				usage();
				exit(0);
			case '-':
				if (i==1) {
				        parse_string(argv[0]+i+1);
                                        i=strlen(argv[0])-1;   /* quit loop */
				}
				break;
			default:
				usage();
        			printf("warning 422: Unknown option \"%c\" ignored.\n",argv[0][i]);
        			exit(1);
			}
			i++;
		}
		argc--;
                argv++;
	}

  argc++; argv--;

  if (RECORD_SIZE > IN_CORE_SIZE) {
printf("record size is set as in core size.\n");
	RECORD_SIZE = IN_CORE_SIZE;
  }

  SgProject * project = new SgProject("test", argc, argv);

  for (i=0; i< project->numberOfFiles(); i++) {

    global_file = &(project->file(i));

    SgStatement * s;

    s = global_file->firstStatement();

    //test_transform(s);
    
    if (!silentFlag)
    printf("\n-------- 'PRE-ANALYSE PHASE' ---------\n"); 
    specification_analyse(s);

    if (!silentFlag)
    printf("-------- 'ANALYSE PHASE' ---------\n"); 
    analyse(s);

    if (!silentFlag)
    printf("-------- 'TRANSFORMATION PHASE' ---------\n"); 

    transform_1(s);
    transform_2(s);
         //slx
//#ifndef USE_F90
//    transform_3(s);	// replace variable by stack address ( for allocate )
//#endif


    char *newfile = newOutputFileName(global_file->filept->filename);
    global_file->saveDepFile(newfile);
//    global_file->filept->filename = newfile;

    free(newfile);

      // free attributes
    for (;s;s=s->lexNext()) 
	s->deleteAllAttribute(); 

    for (SgExpression *exp=global_file->firstExpression(); exp ; exp=exp->nextInExprTable())
	exp->deleteAllAttribute();

    for (SgSymbol *sym=global_file->firstSymbol(); sym ; sym=sym->next())
	sym->deleteAllAttribute();

  };

//    extern void FinalizeTable();
//    FinalizeTable();

  delete project;

  if (debug_trans0) display_comm_feq();
  return 0;
}

